/**
 * @license
 * MIT License
 *
 * Copyright (c) 2018 Goffert van Gool
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global['lite-html'] = {})));
}(this, (function (exports) { 'use strict';

  var marker = Math.random().toString(36).substring(2).padStart(10, '0');
  var attributeMarker = 'font-family:' + marker;
  var IEStyleMarker = 'font-family: ' + marker + ';';
  var commentMarker = 'comment-' + marker;
  var failMarker = 'node-' + marker;
  var nodeMarker = failMarker + '" ' + failMarker + ' ';

  var classCallCheck = function (instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  };

  var createClass = function () {
    function defineProperties(target, props) {
      for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) descriptor.writable = true;
        Object.defineProperty(target, descriptor.key, descriptor);
      }
    }

    return function (Constructor, protoProps, staticProps) {
      if (protoProps) defineProperties(Constructor.prototype, protoProps);
      if (staticProps) defineProperties(Constructor, staticProps);
      return Constructor;
    };
  }();

  var isSerializable = function isSerializable(value) {
    return typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean';
  };
  var isArray = function isArray(nonPrimitive) {
    return Array.isArray(nonPrimitive) || nonPrimitive[Symbol.iterator];
  };

  var noChange = {};
  var ITERABLE = {};

  var NodePart = function () {
    // node OR parent _must_ be defined
    // If a node is defined, this NodePart represents the position of that node in the tree
    // If a only a parent is defined, this NodePart represents the entire content of the parent
    function NodePart(_ref) {
      var node = _ref.node,
          parent = _ref.parent,
          before = _ref.before,
          after = _ref.after;
      classCallCheck(this, NodePart);

      this.iterableParts = [];
      this.node = node;
      this.value = {}; // initial value is an object so any new value will effectively be different

      this.parentNode = parent || node && node.parentNode;
      this.beforeNode = before || node && node.previousSibling;
      this.afterNode = after || node && node.nextSibling;

      this.templateInstances = new Map();
    }

    createClass(NodePart, [{
      key: 'render',
      value: function render(value) {
        if (value === noChange) {
          return;
        }
        if (value == null) {
          this.clear();
        } else if (isSerializable(value)) {
          this._renderText(value);
        } else if (value instanceof TemplateResult) {
          this._renderTemplateResult(value);
        } else if (isArray(value)) {
          this._renderIterable(value);
        } else if (value instanceof Node) {
          this._renderNode(value);
        } else if (value.then !== undefined) {
          this._renderPromise(value);
        } else {
          this._renderText(String(value));
          this.value = String(value);
        }
        // TODO: something smart
        this.value = value;
      }

      /**
       * Render a primitive value in this part
       *
       * Primitive values are rendered as textContent of a TextNode
       */

    }, {
      key: '_renderText',
      value: function _renderText(text) {
        // If the previous value is equal to the primitive, do nothing
        if (this.value === text) {
          return;
        }
        // If the node is a TextNode, replace the content of that node
        // Otherwise, create a new TextNode with the primitive value as content
        if (this.node && this.node.nodeType === 3) {
          this.node.textContent = text;
        } else {
          this._renderNode(document.createTextNode(text));
        }
      }

      /**
       * Render a TemplateResult in this part
       *
       * Checks if this template has already been rendered in this part before.
       * If so, it re-use that TemplateInstance
       * If not, it create a new TemplateInstance
       */

    }, {
      key: '_renderTemplateResult',
      value: function _renderTemplateResult(templateResult) {
        var instance = this.templateInstances.get(templateResult.template);
        if (!instance) {
          instance = new TemplateInstance(templateResult.template, this.parentNode, this.beforeNode, this.afterNode);
          this.templateInstances.set(templateResult.template, instance);
        }
        if (instance.fragment !== this.node) {
          this.clear();
          this.parentNode.insertBefore(instance.fragment, this.afterNode);
          this.node = instance.fragment;
        }
        instance.render(templateResult.values);
      }

      /**
       * Render an iterable in this part
       *
       * Creates a part for each item in the iterable
       * Render each iterable value in a part
       */

    }, {
      key: '_renderIterable',
      value: function _renderIterable(iterable) {
        if (iterable.length === 0) {
          this.clear();
          return;
        }

        if (this.node !== ITERABLE) {
          this.clear();
          this.node = ITERABLE;
          this.iterableParts.length = 0;
        }

        while (this.iterableParts.length < iterable.length) {
          var node = document.createComment('');
          var delimiter = document.createTextNode('');
          this.parentNode.insertBefore(delimiter, this.afterNode);
          this.parentNode.insertBefore(node, delimiter);
          this.iterableParts.push(new NodePart({ node: node, parent: this.parentNode }));
        }

        if (this.iterableParts.length > iterable.length) {
          var nodeToRemove = this.iterableParts[iterable.length].beforeNode.nextSibling;
          moveNodes(this.parentNode, nodeToRemove, this.afterNode);
        }

        this.iterableParts.length = iterable.length;

        this.iterableParts.forEach(function (part, index) {
          part.render(iterable[index]);
        });
      }

      /**
       * Render a DOM node in this part
       */

    }, {
      key: '_renderNode',
      value: function _renderNode(node) {
        // If node is already the current content node, do nothing
        if (this.node === node) {
          return;
        }
        this.clear();
        this.parentNode.insertBefore(node, this.afterNode);
        this.node = node;
      }

      /**
       * Render the result of a promise in this part
       */

    }, {
      key: '_renderPromise',
      value: function _renderPromise(promise) {
        var _this = this;

        // If we rendered this Promise already, do nothing
        // TODO: Account for rendering promise -> other -> promise
        if (this.value === promise || this.promise === promise) {
          return;
        }
        this.clear();
        this.promise = promise;
        this.value = promise;
        // When the promise resolves, render the value of that promise
        promise.then(function (value) {
          // If the part rendered another value in the meantime, abandon the promise
          if (_this.value !== promise) {
            return;
          }
          _this.render(value);
        });
      }

      /**
       * Clear out the content of this NodePart
       *
       * If the current content is part of a DocumentFragment (it is the result of a TemplateResult or an Array)
       * The current content is moved back into that fragment to be used again if the same fragment is rendered
       * Otherwise, the current content is removed from the DOM permanently
       */

    }, {
      key: 'clear',
      value: function clear() {
        var nodeToRemove = this.beforeNode && this.beforeNode.nextSibling || this.parentNode.childNodes[0];
        moveNodes(this.parentNode, nodeToRemove, this.afterNode, this.node);
        this.node = undefined;
        this.value = undefined;
      }
    }]);
    return NodePart;
  }();

  var moveNodes = function moveNodes(parent, begin, end, target) {
    var nextNode = void 0;
    var remove = void 0;
    if (target instanceof DocumentFragment) {
      remove = function remove() {
        return target.appendChild(begin);
      };
    } else {
      remove = function remove() {
        return parent.removeChild(begin);
      };
    }
    while (begin != end) {
      nextNode = begin.nextSibling;
      remove();
      begin = nextNode;
    }
  };

  // The node in the CommentPart constructor must be a CommentNode
  var CommentPart = function () {
    function CommentPart(_ref2) {
      var node = _ref2.node;
      classCallCheck(this, CommentPart);

      this.node = node;
    }

    createClass(CommentPart, [{
      key: 'render',
      value: function render(value) {
        this.node.textContent = value;
      }
    }]);
    return CommentPart;
  }();

  var AttributePart = function () {
    function AttributePart(_ref3) {
      var node = _ref3.node,
          attribute = _ref3.attribute;
      classCallCheck(this, AttributePart);

      this.node = node;
      switch (attribute[0]) {
        case '.':
          this.render = this._renderProperty;
        case '?':
          this.render = this.render || this._renderBoolean;
        case '@':
          this.render = this.render || this._renderEvent;
          this.node.removeAttribute(attribute);
          this.name = attribute.slice(1);
          break;
        default:
          this.render = this._renderAttribute;
          this.name = attribute;
      }
    }

    createClass(AttributePart, [{
      key: '_renderProperty',
      value: function _renderProperty(value) {
        this.node[this.name] = value;
      }
    }, {
      key: '_renderBoolean',
      value: function _renderBoolean(boolean) {
        boolean ? this.node.setAttribute(this.name, '') : this.node.removeAttribute(this.name);
      }
    }, {
      key: '_renderEvent',
      value: function _renderEvent(listener) {
        if (this.value === listener) {
          return;
        }
        this.node.removeEventListener(this.name, this.value);
        this.node.addEventListener(this.name, listener);
        this.value = listener;
      }
    }, {
      key: '_renderAttribute',
      value: function _renderAttribute(string) {
        if (string !== this.value) {
          this.node.setAttribute(this.name, string);
          this.value = string;
        }
      }
    }]);
    return AttributePart;
  }();

  var lastAttributeNameRegex = /[ \x09\x0a\x0c\x0d]([^\0-\x1F\x7F-\x9F \x09\x0a\x0c\x0d"'>=/]+)[ \x09\x0a\x0c\x0d]*=$/;
  var filter = [].filter;

  var findParts = function findParts(strings, template) {
    var parts = [];

    // Recursive depth-first tree traversal that finds nodes in the subtree of `node` that are parts.
    var recursiveIndex = function recursiveIndex(node, path) {
      if (node === undefined) {
        return;
      }
      if (node.nodeType === 8) {
        if (node.nodeValue === commentMarker) {
          parts.push({ type: CommentPart, path: path });
        } else if (node.nodeValue === nodeMarker) {
          parts.push({ type: NodePart, path: path });
        }
        return;
      }

      // Element Node
      if (node.nodeType === 1) {
        if (node.hasAttribute(failMarker)) {
          throw new Error("The '>' character is not allowed in attribute literals. Replace with '&gt;'");
        }
        // All nodes with attribute parts have the attributeMarker set as an attribute
        if (node.hasAttribute(attributeMarker)) {
          node.removeAttribute(attributeMarker);

          var dynamicAttributes = filter.call(node.attributes, function (attribute) {
            return attribute.value === attributeMarker;
          }).length;
          if (node.getAttribute('style') === IEStyleMarker) {
            dynamicAttributes += 1;
          }
          for (var i = 0; i < dynamicAttributes; i++) {
            var attributeMatch = lastAttributeNameRegex.exec(strings[parts.length]);
            var attribute = attributeMatch[1];
            parts.push({ type: AttributePart, path: path, attribute: attribute });
          }
        }
      }
      var children = node.childNodes;
      var length = children.length;
      for (var _i = 0; _i < length; _i++) {
        recursiveIndex(children[_i], path.concat([_i]));
      }
    };

    recursiveIndex(template.content, []);
    if (parts.length < strings.length - 1) {
      throw new Error("Double attribute assignments are not allowed: '<div a=${0} a=${0}>'");
    }
    return parts;
  };

  // The second marker is to add a boolean attribute to the element
  // This is to easily test if a node has dynamic attributes by checking against that attribute
  var attributeMarkerTag = attributeMarker + ' ' + attributeMarker;

  // The space at the end is necessary, to avoid accidentally closing comments with `<!-->`
  var commentMarkerTag = '--><!--' + commentMarker + '--><!-- ';

  // The extra content at the end is to add a flag to an element when
  // a nodeMarkerTag is inserted as an attribute due to an attribute containing `>`
  var nodeMarkerTag = '<!--' + nodeMarker + '-->';

  var attributeContext = Symbol('attributeContext');
  var commentContext = Symbol('commentContext');
  var nodeContext = Symbol('nodeContext');
  var unchangedContext = Symbol('unchangedContext');

  var contextMap = new Map();
  contextMap.set(attributeContext, attributeMarkerTag);
  contextMap.set(commentContext, commentMarkerTag);
  contextMap.set(nodeContext, nodeMarkerTag);

  var parseContext = function parseContext(string) {
    var openComment = string.lastIndexOf('<!--');
    var closeComment = string.indexOf('-->', openComment + 1);
    var commentClosed = closeComment > -1;
    var type = void 0;
    if (openComment > -1 && !commentClosed) {
      type = commentContext;
    } else {
      var closeTag = string.lastIndexOf('>');
      var openTag = string.indexOf('<', closeTag + 1);
      if (openTag > -1) {
        type = attributeContext;
      } else {
        if (closeTag > -1) {
          type = nodeContext;
        } else {
          type = unchangedContext;
        }
      }
    }
    return { commentClosed: commentClosed, type: type };
  };

  var parseTemplate = function parseTemplate(strings) {
    var html = [];
    var lastStringIndex = strings.length - 1;
    var currentContext = nodeContext;
    for (var i = 0; i < lastStringIndex; i++) {
      var string = strings[i];
      var context = parseContext(string);
      if ((currentContext !== commentContext || context.commentClosed) && context.type !== unchangedContext) {
        currentContext = context.type;
      }
      if (currentContext === attributeContext && string.slice(-1) !== '=') {
        throw new Error('Only bare attribute parts are allowed: `<div a=${0}>`');
      }
      html.push(string);
      html.push(contextMap.get(currentContext));
    }

    html.push(strings[lastStringIndex]);
    return html.join('');
  };

  var buildTemplate = function buildTemplate(strings) {
    var template = document.createElement('template');
    template.innerHTML = parseTemplate(strings);
    return template;
  };

  /**
   * A map that contains all the template literals we have seen before
   * It maps from a String array to a Template object
   *
   * @typedef {Map.<[String], Template>}
   */
  var templateMap = new Map();

  /**
   * Template holds the DocumentFragment that is to be used as a prototype for instances of this template
   * When a template is to be rendered in a new location, a clone will be made from this
   *
   * @prop {[String]} strings
   *   The unique string array that this template represents
   * @prop {[DocumentFragment]} element
   *   The DocumentFragment that can be cloned to make instances of this template
   * @prop {[Object]} parts
   *   The descriptions of the parts in this Template. Each part has a path which defines a unique location in the
   *   template DOM tree, a type which defines the part type, and an optional attribute which defines the name of
   *   the attribute this part represents.
   */
  var Template = function Template(strings) {
    classCallCheck(this, Template);

    this.strings = strings;
    this.element = buildTemplate(strings);
    this.parts = findParts(strings, this.element);
  };

  /**
   * TemplateResult holds the strings and values that result from a tagged template string literal.
   * TemplateResult can find and return a unique Template object that represents its tagged template string literal.
   */
  var TemplateResult = function () {
    function TemplateResult(strings, values) {
      classCallCheck(this, TemplateResult);

      this.strings = strings;
      this.values = values;
      this._template = undefined;
    }

    createClass(TemplateResult, [{
      key: 'template',
      get: function get$$1() {
        if (this._template) {
          return this._template;
        }
        var template = templateMap.get(this.strings);
        if (!template) {
          template = new Template(this.strings);
          templateMap.set(this.strings, template);
        }
        this._template = template;
        return template;
      }
    }]);
    return TemplateResult;
  }();

  /**
   * An instance of a template that can be rendered somewhere
   *
   * @prop {Template} template
   *   The unique Template object that this is an instance of
   * @prop {[DocumentFragment]} fragment
   *   The DocumentFragment that is a clone of the Template's prototype DocumentFragment
   * @prop {[AttributePart|CommentPart|NodePart|]} parts
   *   The parts that render into this template instance
   */
  var TemplateInstance = function () {
    function TemplateInstance(template, parent, before, after) {
      var _this = this;

      classCallCheck(this, TemplateInstance);

      this.template = template;
      this.fragment = template.element.content.cloneNode(true);

      // Create new Parts based on the part definitions set on the Template
      var parts = this.template.parts.map(function (part) {
        var node = _this.fragment;
        part.path.forEach(function (nodeIndex) {
          node = node.childNodes[nodeIndex];
        });
        part.node = node;
        if (part.type === NodePart) {
          if (part.path.length === 1) {
            part.parent = parent;
            part.before = node.previousSibling || before;
            part.after = node.nextSibling || after;
          } else {
            part.parent = node.parentNode;
          }
        }
        return part;
      });
      this.parts = parts.map(function (part) {
        return new part.type(part);
      });
    }

    /**
     * Render values into the parts of this TemplateInstance
     *
     * @param {[any]} values
     *   An array of values to render into the parts. There should be one value per part
     */

    createClass(TemplateInstance, [{
      key: 'render',
      value: function render(values) {
        this.parts.map(function (part, index) {
          return part.render(values[index]);
        });
      }
    }]);
    return TemplateInstance;
  }();

  /**
   * Tagging function to tag JavaScript template string literals as HTML
   *
   * @return {TemplateResult}
   *   The strings and values of the template string wrapped in a TemplateResult object
   */
  var html = function html(strings) {
    for (var _len = arguments.length, values = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      values[_key - 1] = arguments[_key];
    }

    return new TemplateResult(strings, values);
  };

  /**
   * Render content into a target node
   *
   * @param {any} content
   *   Any content you wish to render. Usually a template string literal tagged with the `html` function
   * @param {Node} target
   *   An HTML Node that you wish to render the content into.
   *   The content will become the sole content of the target node.
   */
  var render = function render(content, target) {
    // Check if the target has a NodePart that represents its content
    var part = target.__nodePart;
    if (!part) {
      // If it does not, create a new NodePart
      part = new NodePart({ parent: target });
      target.__nodePart = part;
    }
    // Task the NodePart of this target to render the content
    part.render(content);
  };

  exports.html = html;
  exports.render = render;
  exports.noChange = noChange;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=lite-html.es5.js.map
