/**
 * @license
 * MIT License
 *
 * Copyright (c) 2018 Goffert van Gool
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

const marker = Math.random()
  .toString(36)
  .substring(2)
  .padStart(10, '0');
const attributeMarker = `font-family:${marker}`;
const IEStyleMarker = `font-family: ${marker};`;
const commentMarker = `comment-${marker}`;
const failMarker = `node-${marker}`;
const nodeMarker = `${failMarker}" ${failMarker} `;

const isSerializable = value => typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean';
const isArray = nonPrimitive => Array.isArray(nonPrimitive) || nonPrimitive[Symbol.iterator];

const noChange = {};
const ITERABLE = {};

class NodePart {
  // node OR parent _must_ be defined
  // If a node is defined, this NodePart represents the position of that node in the tree
  // If a only a parent is defined, this NodePart represents the entire content of the parent
  constructor({ node, parent, before, after }) {
    this.iterableParts = [];
    this.node = node;
    this.value = {}; // initial value is an object so any new value will effectively be different

    this.parentNode = parent || (node && node.parentNode);
    this.beforeNode = before || (node && node.previousSibling);
    this.afterNode = after || (node && node.nextSibling);

    this.templateInstances = new Map();
  }

  render(value) {
    if (value === noChange) {
      return;
    }
    if (value == null) {
      this.clear();
    } else if (isSerializable(value)) {
      this._renderText(value);
    } else if (value instanceof TemplateResult) {
      this._renderTemplateResult(value);
    } else if (isArray(value)) {
      this._renderIterable(value);
    } else if (value instanceof Node) {
      this._renderNode(value);
    } else if (value.then !== undefined) {
      this._renderPromise(value);
    } else {
      this._renderText(String(value));
      this.value = String(value);
    }
    // TODO: something smart
    this.value = value;
  }

  /**
   * Render a primitive value in this part
   *
   * Primitive values are rendered as textContent of a TextNode
   */
  _renderText(text) {
    // If the previous value is equal to the primitive, do nothing
    if (this.value === text) {
      return;
    }
    // If the node is a TextNode, replace the content of that node
    // Otherwise, create a new TextNode with the primitive value as content
    if (this.node && this.node.nodeType === 3) {
      this.node.textContent = text;
    } else {
      this._renderNode(document.createTextNode(text));
    }
  }

  /**
   * Render a TemplateResult in this part
   *
   * Checks if this template has already been rendered in this part before.
   * If so, it re-use that TemplateInstance
   * If not, it create a new TemplateInstance
   */
  _renderTemplateResult(templateResult) {
    let instance = this.templateInstances.get(templateResult.template);
    if (!instance) {
      instance = new TemplateInstance(templateResult.template, this.parentNode, this.beforeNode, this.afterNode);
      this.templateInstances.set(templateResult.template, instance);
    }
    if (instance.fragment !== this.node) {
      this.clear();
      this.parentNode.insertBefore(instance.fragment, this.afterNode);
      this.node = instance.fragment;
    }
    instance.render(templateResult.values);
  }

  /**
   * Render an iterable in this part
   *
   * Creates a part for each item in the iterable
   * Render each iterable value in a part
   */
  _renderIterable(iterable) {
    if (iterable.length === 0) {
      this.clear();
      return;
    }

    if (this.node !== ITERABLE) {
      this.clear();
      this.node = ITERABLE;
      this.iterableParts.length = 0;
    }

    while (this.iterableParts.length < iterable.length) {
      const node = document.createComment('');
      const delimiter = document.createTextNode('');
      this.parentNode.insertBefore(delimiter, this.afterNode);
      this.parentNode.insertBefore(node, delimiter);
      this.iterableParts.push(new NodePart({ node, parent: this.parentNode }));
    }

    if (this.iterableParts.length > iterable.length) {
      let nodeToRemove = this.iterableParts[iterable.length].beforeNode.nextSibling;
      moveNodes(this.parentNode, nodeToRemove, this.afterNode);
    }

    this.iterableParts.length = iterable.length;

    this.iterableParts.forEach((part, index) => {
      part.render(iterable[index]);
    });
  }

  /**
   * Render a DOM node in this part
   */
  _renderNode(node) {
    // If node is already the current content node, do nothing
    if (this.node === node) {
      return;
    }
    this.clear();
    this.parentNode.insertBefore(node, this.afterNode);
    this.node = node;
  }

  /**
   * Render the result of a promise in this part
   */
  _renderPromise(promise) {
    // If we rendered this Promise already, do nothing
    // TODO: Account for rendering promise -> other -> promise
    if (this.value === promise || this.promise === promise) {
      return;
    }
    this.clear();
    this.promise = promise;
    this.value = promise;
    // When the promise resolves, render the value of that promise
    promise.then(value => {
      // If the part rendered another value in the meantime, abandon the promise
      if (this.value !== promise) {
        return;
      }
      this.render(value);
    });
  }

  /**
   * Clear out the content of this NodePart
   *
   * If the current content is part of a DocumentFragment (it is the result of a TemplateResult or an Array)
   * The current content is moved back into that fragment to be used again if the same fragment is rendered
   * Otherwise, the current content is removed from the DOM permanently
   */
  clear() {
    let nodeToRemove = (this.beforeNode && this.beforeNode.nextSibling) || this.parentNode.childNodes[0];
    moveNodes(this.parentNode, nodeToRemove, this.afterNode, this.node);
    this.node = undefined;
    this.value = undefined;
  }
}

const moveNodes = (parent, begin, end, target) => {
  let nextNode;
  let remove;
  if (target instanceof DocumentFragment) {
    remove = () => target.appendChild(begin);
  } else {
    remove = () => parent.removeChild(begin);
  }
  while (begin != end) {
    nextNode = begin.nextSibling;
    remove();
    begin = nextNode;
  }
};

// The node in the CommentPart constructor must be a CommentNode
class CommentPart {
  constructor({ node }) {
    this.node = node;
  }

  render(value) {
    this.node.textContent = value;
  }
}

class AttributePart {
  constructor({ node, attribute }) {
    this.node = node;
    switch (attribute[0]) {
      case '.':
        this.render = this._renderProperty;
      case '?':
        this.render = this.render || this._renderBoolean;
      case '@':
        this.render = this.render || this._renderEvent;
        this.node.removeAttribute(attribute);
        this.name = attribute.slice(1);
        break;
      default:
        this.render = this._renderAttribute;
        this.name = attribute;
    }
  }

  _renderProperty(value) {
    this.node[this.name] = value;
  }

  _renderBoolean(boolean) {
    boolean ? this.node.setAttribute(this.name, '') : this.node.removeAttribute(this.name);
  }

  _renderEvent(listener) {
    if (this.value === listener) {
      return;
    }
    this.node.removeEventListener(this.name, this.value);
    this.node.addEventListener(this.name, listener);
    this.value = listener;
  }

  _renderAttribute(string) {
    if (string !== this.value) {
      this.node.setAttribute(this.name, string);
      this.value = string;
    }
  }
}

const lastAttributeNameRegex = /[ \x09\x0a\x0c\x0d]([^\0-\x1F\x7F-\x9F \x09\x0a\x0c\x0d"'>=/]+)[ \x09\x0a\x0c\x0d]*=$/;
const filter = [].filter;

const findParts = (strings, template) => {
  let parts = [];

  // Recursive depth-first tree traversal that finds nodes in the subtree of `node` that are parts.
  const recursiveIndex = (node, path) => {
    if (node === undefined) {
      return;
    }
    if (node.nodeType === 8) {
      if (node.nodeValue === commentMarker) {
        parts.push({ type: CommentPart, path });
      } else if (node.nodeValue === nodeMarker) {
        parts.push({ type: NodePart, path });
      }
      return;
    }

    // Element Node
    if (node.nodeType === 1) {
      if (node.hasAttribute(failMarker)) {
        throw new Error("The '>' character is not allowed in attribute literals. Replace with '&gt;'");
      }
      // All nodes with attribute parts have the attributeMarker set as an attribute
      if (node.hasAttribute(attributeMarker)) {
        node.removeAttribute(attributeMarker);

        let dynamicAttributes = filter.call(node.attributes, attribute => attribute.value === attributeMarker).length;
        if (node.getAttribute('style') === IEStyleMarker) {
          dynamicAttributes += 1;
        }
        for (let i = 0; i < dynamicAttributes; i++) {
          const attributeMatch = lastAttributeNameRegex.exec(strings[parts.length]);
          const attribute = attributeMatch[1];
          parts.push({ type: AttributePart, path, attribute });
        }
      }
    }
    const children = node.childNodes;
    const length = children.length;
    for (let i = 0; i < length; i++) {
      recursiveIndex(children[i], path.concat([i]));
    }
  };

  recursiveIndex(template.content, []);
  if (parts.length < strings.length - 1) {
    throw new Error("Double attribute assignments are not allowed: '<div a=${0} a=${0}>'");
  }
  return parts;
};

// The second marker is to add a boolean attribute to the element
// This is to easily test if a node has dynamic attributes by checking against that attribute
const attributeMarkerTag = `${attributeMarker} ${attributeMarker}`;

// The space at the end is necessary, to avoid accidentally closing comments with `<!-->`
const commentMarkerTag = `--><!--${commentMarker}--><!-- `;

// The extra content at the end is to add a flag to an element when
// a nodeMarkerTag is inserted as an attribute due to an attribute containing `>`
const nodeMarkerTag = `<!--${nodeMarker}-->`;

const attributeContext = Symbol('attributeContext');
const commentContext = Symbol('commentContext');
const nodeContext = Symbol('nodeContext');
const unchangedContext = Symbol('unchangedContext');

const contextMap = new Map();
contextMap.set(attributeContext, attributeMarkerTag);
contextMap.set(commentContext, commentMarkerTag);
contextMap.set(nodeContext, nodeMarkerTag);

const parseContext = string => {
  const openComment = string.lastIndexOf('<!--');
  const closeComment = string.indexOf('-->', openComment + 1);
  const commentClosed = closeComment > -1;
  let type;
  if (openComment > -1 && !commentClosed) {
    type = commentContext;
  } else {
    const closeTag = string.lastIndexOf('>');
    const openTag = string.indexOf('<', closeTag + 1);
    if (openTag > -1) {
      type = attributeContext;
    } else {
      if (closeTag > -1) {
        type = nodeContext;
      } else {
        type = unchangedContext;
      }
    }
  }
  return { commentClosed, type };
};

const parseTemplate = strings => {
  const html = [];
  const lastStringIndex = strings.length - 1;
  let currentContext = nodeContext;
  for (let i = 0; i < lastStringIndex; i++) {
    const string = strings[i];
    const context = parseContext(string);
    if ((currentContext !== commentContext || context.commentClosed) && context.type !== unchangedContext) {
      currentContext = context.type;
    }
    if (currentContext === attributeContext && string.slice(-1) !== '=') {
      throw new Error('Only bare attribute parts are allowed: `<div a=${0}>`');
    }
    html.push(string);
    html.push(contextMap.get(currentContext));
  }

  html.push(strings[lastStringIndex]);
  return html.join('');
};

const buildTemplate = strings => {
  const template = document.createElement('template');
  template.innerHTML = parseTemplate(strings);
  return template;
};

/**
 * A map that contains all the template literals we have seen before
 * It maps from a String array to a Template object
 *
 * @typedef {Map.<[String], Template>}
 */
const templateMap = new Map();

/**
 * Template holds the DocumentFragment that is to be used as a prototype for instances of this template
 * When a template is to be rendered in a new location, a clone will be made from this
 *
 * @prop {[String]} strings
 *   The unique string array that this template represents
 * @prop {[DocumentFragment]} element
 *   The DocumentFragment that can be cloned to make instances of this template
 * @prop {[Object]} parts
 *   The descriptions of the parts in this Template. Each part has a path which defines a unique location in the
 *   template DOM tree, a type which defines the part type, and an optional attribute which defines the name of
 *   the attribute this part represents.
 */
class Template {
  constructor(strings) {
    this.strings = strings;
    this.element = buildTemplate(strings);
    this.parts = findParts(strings, this.element);
  }
}

/**
 * TemplateResult holds the strings and values that result from a tagged template string literal.
 * TemplateResult can find and return a unique Template object that represents its tagged template string literal.
 */
class TemplateResult {
  constructor(strings, values) {
    this.strings = strings;
    this.values = values;
    this._template = undefined;
  }

  get template() {
    if (this._template) {
      return this._template;
    }
    let template = templateMap.get(this.strings);
    if (!template) {
      template = new Template(this.strings);
      templateMap.set(this.strings, template);
    }
    this._template = template;
    return template;
  }
}

/**
 * An instance of a template that can be rendered somewhere
 *
 * @prop {Template} template
 *   The unique Template object that this is an instance of
 * @prop {[DocumentFragment]} fragment
 *   The DocumentFragment that is a clone of the Template's prototype DocumentFragment
 * @prop {[AttributePart|CommentPart|NodePart|]} parts
 *   The parts that render into this template instance
 */
class TemplateInstance {
  constructor(template, parent, before, after) {
    this.template = template;
    this.fragment = template.element.content.cloneNode(true);

    // Create new Parts based on the part definitions set on the Template
    const parts = this.template.parts.map(part => {
      let node = this.fragment;
      part.path.forEach(nodeIndex => {
        node = node.childNodes[nodeIndex];
      });
      part.node = node;
      if (part.type === NodePart) {
        if (part.path.length === 1) {
          part.parent = parent;
          part.before = node.previousSibling || before;
          part.after = node.nextSibling || after;
        } else {
          part.parent = node.parentNode;
        }
      }
      return part;
    });
    this.parts = parts.map(part => new part.type(part));
  }

  /**
   * Render values into the parts of this TemplateInstance
   *
   * @param {[any]} values
   *   An array of values to render into the parts. There should be one value per part
   */
  render(values) {
    this.parts.map((part, index) => part.render(values[index]));
  }
}

/**
 * Tagging function to tag JavaScript template string literals as HTML
 *
 * @return {TemplateResult}
 *   The strings and values of the template string wrapped in a TemplateResult object
 */
const html = (strings, ...values) => {
  return new TemplateResult(strings, values);
};

/**
 * Render content into a target node
 *
 * @param {any} content
 *   Any content you wish to render. Usually a template string literal tagged with the `html` function
 * @param {Node} target
 *   An HTML Node that you wish to render the content into.
 *   The content will become the sole content of the target node.
 */
const render = (content, target) => {
  // Check if the target has a NodePart that represents its content
  let part = target.__nodePart;
  if (!part) {
    // If it does not, create a new NodePart
    part = new NodePart({ parent: target });
    target.__nodePart = part;
  }
  // Task the NodePart of this target to render the content
  part.render(content);
};

export { html, render, noChange };
//# sourceMappingURL=lite-html.js.map
