/**
 * @license
 * MIT License
 *
 * Copyright (c) 2018 Goffert van Gool
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

import { TemplateResult, TemplateInstance } from './templates.js';

export const isSerializable = value => typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean';
export const isArray = nonPrimitive => Array.isArray(nonPrimitive) || nonPrimitive[Symbol.iterator];

export const noChange = {};
const ITERABLE = {};

export class NodePart {
  // node OR parent _must_ be defined
  // If a node is defined, this NodePart represents the position of that node in the tree
  // If a only a parent is defined, this NodePart represents the entire content of the parent
  constructor({ node, parent, before, after }) {
    this.iterableParts = [];
    this.node = node;
    this.value = {}; // initial value is an object so any new value will effectively be different

    this.parentNode = parent || (node && node.parentNode);
    this.beforeNode = before || (node && node.previousSibling);
    this.afterNode = after || (node && node.nextSibling);

    this.templateInstances = new Map();
  }

  render(value) {
    if (value === noChange) {
      return;
    }
    if (value == null) {
      this.clear();
    } else if (isSerializable(value)) {
      this._renderText(value);
    } else if (value instanceof TemplateResult) {
      this._renderTemplateResult(value);
    } else if (isArray(value)) {
      this._renderIterable(value);
    } else if (value instanceof Node) {
      this._renderNode(value);
    } else if (value.then !== undefined) {
      this._renderPromise(value);
    } else {
      this._renderText(String(value));
      this.value = String(value);
    }
    // TODO: something smart
    this.value = value;
  }

  /**
   * Render a primitive value in this part
   *
   * Primitive values are rendered as textContent of a TextNode
   */
  _renderText(text) {
    // If the previous value is equal to the primitive, do nothing
    if (this.value === text) {
      return;
    }
    // If the node is a TextNode, replace the content of that node
    // Otherwise, create a new TextNode with the primitive value as content
    if (this.node && this.node.nodeType === 3) {
      this.node.textContent = text;
    } else {
      this._renderNode(document.createTextNode(text));
    }
  }

  /**
   * Render a TemplateResult in this part
   *
   * Checks if this template has already been rendered in this part before.
   * If so, it re-use that TemplateInstance
   * If not, it create a new TemplateInstance
   */
  _renderTemplateResult(templateResult) {
    let instance = this.templateInstances.get(templateResult.template);
    if (!instance) {
      instance = new TemplateInstance(templateResult.template, this.parentNode, this.beforeNode, this.afterNode);
      this.templateInstances.set(templateResult.template, instance);
    }
    if (instance.fragment !== this.node) {
      this.clear();
      this.parentNode.insertBefore(instance.fragment, this.afterNode);
      this.node = instance.fragment;
    }
    instance.render(templateResult.values);
  }

  /**
   * Render an iterable in this part
   *
   * Creates a part for each item in the iterable
   * Render each iterable value in a part
   */
  _renderIterable(iterable) {
    if (iterable.length === 0) {
      this.clear();
      return;
    }

    if (this.node !== ITERABLE) {
      this.clear();
      this.node = ITERABLE;
      this.iterableParts.length = 0;
    }

    while (this.iterableParts.length < iterable.length) {
      const node = document.createComment('');
      const delimiter = document.createTextNode('');
      this.parentNode.insertBefore(delimiter, this.afterNode);
      this.parentNode.insertBefore(node, delimiter);
      this.iterableParts.push(new NodePart({ node, parent: this.parentNode }));
    }

    if (this.iterableParts.length > iterable.length) {
      let nodeToRemove = this.iterableParts[iterable.length].beforeNode.nextSibling;
      moveNodes(this.parentNode, nodeToRemove, this.afterNode);
    }

    this.iterableParts.length = iterable.length;

    this.iterableParts.forEach((part, index) => {
      part.render(iterable[index]);
    });
  }

  /**
   * Render a DOM node in this part
   */
  _renderNode(node) {
    // If node is already the current content node, do nothing
    if (this.node === node) {
      return;
    }
    this.clear();
    this.parentNode.insertBefore(node, this.afterNode);
    this.node = node;
  }

  /**
   * Render the result of a promise in this part
   */
  _renderPromise(promise) {
    // If we rendered this Promise already, do nothing
    // TODO: Account for rendering promise -> other -> promise
    if (this.value === promise || this.promise === promise) {
      return;
    }
    this.clear();
    this.promise = promise;
    this.value = promise;
    // When the promise resolves, render the value of that promise
    promise.then(value => {
      // If the part rendered another value in the meantime, abandon the promise
      if (this.value !== promise) {
        return;
      }
      this.render(value);
    });
  }

  /**
   * Clear out the content of this NodePart
   *
   * If the current content is part of a DocumentFragment (it is the result of a TemplateResult or an Array)
   * The current content is moved back into that fragment to be used again if the same fragment is rendered
   * Otherwise, the current content is removed from the DOM permanently
   */
  clear() {
    let nodeToRemove = (this.beforeNode && this.beforeNode.nextSibling) || this.parentNode.childNodes[0];
    moveNodes(this.parentNode, nodeToRemove, this.afterNode, this.node);
    this.node = undefined;
    this.value = undefined;
  }
}

const moveNodes = (parent, begin, end, target) => {
  let nextNode;
  let remove;
  if (target instanceof DocumentFragment) {
    remove = () => target.appendChild(begin);
  } else {
    remove = () => parent.removeChild(begin);
  }
  while (begin != end) {
    nextNode = begin.nextSibling;
    remove();
    begin = nextNode;
  }
};

// The node in the CommentPart constructor must be a CommentNode
export class CommentPart {
  constructor({ node }) {
    this.node = node;
  }

  render(value) {
    this.node.textContent = value;
  }
}

export class AttributePart {
  constructor({ node, attribute }) {
    this.node = node;
    switch (attribute[0]) {
      case '.':
        this.render = this._renderProperty;
      case '?':
        this.render = this.render || this._renderBoolean;
      case '@':
        this.render = this.render || this._renderEvent;
        this.node.removeAttribute(attribute);
        this.name = attribute.slice(1);
        break;
      default:
        this.render = this._renderAttribute;
        this.name = attribute;
    }
  }

  _renderProperty(value) {
    this.node[this.name] = value;
  }

  _renderBoolean(boolean) {
    boolean ? this.node.setAttribute(this.name, '') : this.node.removeAttribute(this.name);
  }

  _renderEvent(listener) {
    if (this.value === listener) {
      return;
    }
    this.node.removeEventListener(this.name, this.value);
    this.node.addEventListener(this.name, listener);
    this.value = listener;
  }

  _renderAttribute(string) {
    if (string !== this.value) {
      this.node.setAttribute(this.name, string);
      this.value = string;
    }
  }
}
